#ifndef _BENCH_H
#define _BENCH_H
/** Semantics of DUMP annotations
 * 1. A DUMP* statement forces the analyzer to print the
 *    inferred interval for the variable at the current
 *    program point.
 * 2. DUMP* statements are identified by their line numbers
 *    (since few other informations are shared between 
 *    different analyzers). Therefore, each DUMP must
 *    be put on a separate line.
 * 3. DUMP* statements should never influence the possible
 *    values of the variable.
 * 4. DUMP* statements can only be used to print integer
 *    variables. Signedness information is important:
 *    unsigned variables must be printed using DUMPU*
 *    statements.
 * 
 * There are several variants of the DUMP statement:
 * - DUMPN(n, x) : dumps the inferred interval for signed variable x,
 *                 printed using the name n. For instance,
 *                 DUMPN(x_1, x) will print "##42: x_1: [...]".
 *                 n must be a valid C identifier
 *                 (it is not a string constant).
 * - DUMPNU(n, x) : DUMPN variant for unsigned variables. It
 *                  adds a "U_" prefix to the variable name,
 *                  that is later used to identify it as an unsigned
 *                  variable.
 * - DUMP(x) : an alias for DUMPN(x, x). Generic variant.
 * - DUMPU(x) : unsigned version of DUMP.
 * - DUMPL/DUMPLU : variants for loop iteration variables ("LI_" prefix)
 * - DUMPE/DUMPEU : variants for loop exits ("LE_" prefix)
 * - DUMPP/DUMPPU : variants for function parameters ("PA_" prefix)
 * - DUMPR/DUMPRU : variants for return values ("RV_" prefix)
 */

// Common code for all compilers/analyzers

// Preprocessor hack to include line numbers in the macros
#define STRINGIFY(x) #x
#define STRINGIFY_(x) STRINGIFY(x)
#define S__LINE__ STRINGIFY_(__LINE__)

// This #if identifies when we are compiling with GCC only,
// (and not using Frama-C).
#if defined(__GNUC__) && !defined(__FRAMA_C__) && !defined(__clang__) /** GCC */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>

// When compiling just with GCC, we ignore the annotations,
// unless the special flag PRINT_DUMP_LINES is defined. This
// indicates we should dump line numbers for each DUMP statement.
// This is useful to compare the outputs of the different analyzers.
# ifdef PRINT_DUMP_LINES
#  define DUMPN(n, x) (__LINE__:DUMP)
# else
#  define DUMPN(n, x)
# endif

#elif defined(__COMPCERT__) /** CompCert */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>

int __builtin_nondet_int() {
  int i, j;
  if (i) {
    if (j) {
      i = -2147483648;
    } else {
      i = 2147483647;
    }
  } else {
    if (j) {
      i = 0;
    } else {
      i = 4294967295;
    }
  }
  return i;
}
int nondet_int() { return __builtin_nondet_int(); }

void __dump(const char *s, int n) { int i = n; }

#define DUMPN(n, x) (__dump("##" S__LINE__ ": " # n " = %s", x))

#elif defined(__clang__) /** CLANG (Wrapped) */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>

// This hack allows us to easily associate program variables to line numbers,
// since otherwise Wrapped's variable names do not allow us to retrieve the
// intervals we need
#define ___DUMPN(l, x) do { int l = x; if (l+0) { l = x+0; }; printf("%d", l); } while (0)
#define __DUMPN(line, n, x) ___DUMPN(_dump_LINE_ ## line ## _VAR_ ## n, x)
#define _DUMPN(line, n, x) __DUMPN(line, n, x)
#define DUMPN(n, x) _DUMPN(__LINE__, n, x)

#else /** FRAMA-C */

// We use Frama-C's annotation mechanism
// Note: this mechanism prints variable values at each passage
// during the fixpoint iteration, so scripts are later needed to
// retrieve the latest information. This prevens the usage of
// the -slevel argument, otherwise we will not be able to differentiate
// between different contexts and different passages of the iterator.
#define FCSN(x, y, z) (Frama_C_show_each_LINE_ ## x ## _ ## y(z))
#define FCSN_(x, y, z) FCSN(x, y, z)
#define S_FCSN(x, y) FCSN_(__LINE__, x, y)
#define DUMPN(name, x) S_FCSN(name, x)

#include "/usr/share/frama-c/libc/ctype.h"
#include "/usr/share/frama-c/libc/stdio.h"
#include "/usr/share/frama-c/libc/stdlib.h"
#include "/usr/share/frama-c/libc/string.h"
#include "/usr/share/frama-c/libc/math.h"
#include "/usr/share/frama-c/libc/limits.h"
#include "/usr/share/frama-c/libc/assert.h"

// Some constants are defined here, since we cannot include the library headers
// due to conflicts with redefinitions
#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef HUGE_VAL
#define HUGE_VAL 0x7ff0000000000000ULL
#endif

//#include "/usr/share/frama-c/nondet.h"
#define DBL_EPSILON 2.2204460492503131e-16


#endif /*defined(__GNUC__) && !defined(__FRAMA_C__)*/

/** Common code */
// This code is used by all analyzers. Now that DUMPN
// is defined according to each analyzer, we can define its variants
// once and for all.

#define DUMPNU(n, x) DUMPN(U_ ## n, x)
#define DUMP(x) DUMPN(x, x)
#define DUMPU(x) DUMPNU(x, x)
#define DUMPL(x) DUMPN(LI_ ## x, x)
#define DUMPLU(x) DUMPNU(LI_ ## x, x)
#define DUMPE(x) DUMPN(LE_ ## x, x)
#define DUMPEU(x) DUMPNU(LE_ ## x, x)
#define DUMPP(x) DUMPN(PA_ ## x, x)
#define DUMPPU(x) DUMPNU(PA_ ## x, x)
#define DUMPR(x) DUMPN(RV_ ## x, x)
#define DUMPRU(x) DUMPNU(RV_ ## x, x)

#endif /*_BENCH_H*/
