#!/usr/bin/python
# -*- coding: latin-1 -*-

# Syntax: $0 <file.value.cfg>

import sys
import re

"""Tries to match string s in regex re and retrieve all groups.
Prints a useful message in case of failure.
Returns a tuple containing captured groups.
"""
def match(re, s):
  m = re.search(s)
  if m is None:
    raise Exception("Regex assertion error: re = %s, s = %s" % (re.pattern, s))
  return m.groups()

def normalize_hexa(s):
  try:
    if "0x" in s.lower():
      return str(int(s,16))
    else:
      return str(int(s,10))
  except ValueError:
    # Not an integer: maybe a +oo, or T, or Bot...
    return s

re_interval_str = "\[([^,]+),([^]]+)\]"
re_interval = re.compile(re_interval_str)


def normalize_hexa_itv(itv_str):
  global re_interval
  [(imin, imax)] = re_interval.findall(itv_str)
  return "[%s,%s]" % (normalize_hexa(imin), normalize_hexa(imax))

"""
Searches the CFG value analysis dump for an expression representing a
DUMP annotation and parses it (and the following lines) to obtain
the interval for the DUMPed variable.

New: since intervals now may have two different representations,
     we check the kind of the DUMP (DUMP or DUMPU) and use it
     to choose the corresponding interval.
"""

# Regex for the string containing the DUMP
re_dump_str = "__stringlit_\d+\(\"##(\d+): \"?(\w+)\"? = \%s"
re_dump = re.compile(re_dump_str)

re_intervals_str = "<s([^|]+)\|u([^>]+)>"
re_intervals = re.compile(re_intervals_str)

# Regex for the line containing the value analysis.
# This regex depends on the index of the desired variable,
# which must be used to replace the "%s".
re_intv_str = "<%s>:(\S*)"

# Regex for the assignment that allows to identify the variable to be printed
re_assign_str = "\d+: <\d+> = <(\d+)>"
re_assign = re.compile(re_assign_str)

filename = sys.argv[1]
lines = [line.strip() for line in open(filename)]
i = 0
n = len(lines)

while i < n:
  m = re_dump.search(lines[i])
  if m is None:
    # Nothing interesting
    i += 1
    continue
  # Found an annotation
  line_number = m.group(1)
  var_name = m.group(2)
  is_signed = var_name.strip()[0:2] != "U_"
  # Note: we do not remove the "U_" prefix to be able to compare it with the
  #       other programs

  # Line i+2 should contain an assignment
  (var_number,) = match(re_assign, lines[i+2])
  # Line i+1 should contain an interval for the required variable
  try:
    re_int = re.compile(re_intv_str % var_number)
    (interval,) = match(re_int, lines[i+1])
    # We check if it's a "dual" interval
    m = re_intervals.search(interval)
    if m is None:
      # It is a "single" interval
      pass
    else:
      # We choose the interval corresponding to the DUMP
      #print "[debug] entire interval: ", interval
      interval = m.group(1) if is_signed else m.group(2)
  except Exception:
    # Regex failed: the analysis possibly has a "bottom"
    if "Bot" in lines[i+1]:
      interval = "Bot"
    else:
      # Variable has not been printed: it contains Top
      if is_signed:
        interval = "[-oo,+oo]"
      else:
        interval = "[0,0xFFFFFFFF]"
  # We replace hexadecimal numbers with decimal ones
  if "0x" in interval.lower():
    interval = normalize_hexa_itv(interval)
  # Printing the result
  print "##%s: %s = %s" % (line_number, var_name, interval)
  i += 1

