#!/usr/bin/python
# -*- coding: latin-1 -*-

# Intervals library
# Parses and compares intervals for automatic benchmark analysis

import re

MIN_INT = -2147483648
MAX_INT = 2147483647
MAX_UINT = 4294967295
HALF_LENGTH = 2147483648

modulo_re = re.compile("^(.*) \((\d+)%(\d+)\)")

# Converts a numeric string (or +oo/-oo) into its integer representation
def itv_int_to_int(int_str):
  if int_str == "+oo":
    return MAX_INT
  elif int_str == "-oo":
    return MIN_INT
  elif "|" in int_str:
    # A "dual" interval from Wrapped; we choose the signed value
    int_str = int_str[int_str.rfind(":")+1:]
  return int(int_str)

# Converts an 'interval string' into a triple (min, max, warnings).
# An 'interval string' is a numeric string possibly containing
# an "infinity" symbol.
# warnings: specify whether there has been a loss of precision
# due to the use of k-sets.
def itv_parse(itv_str):
  modulo = 1
  itv_str = itv_str.strip()
  if (itv_str[0] == '['):
    # Interval
    # First we check if there is a "modulo" information
    warnings = []
    if "%" in itv_str:
      m = modulo_re.search(itv_str)
      # We are more interested in the "divisor" than the "remainder"
      mod_remainder = m.group(2)
      modulo = int(m.group(3))
      # We erase the modulo information from the string
      itv_str = m.group(1)
    itv_str = itv_str.strip("[]")
    separator = ".." if ".." in itv_str else ","
    [min_str, max_str] = itv_str.split(separator)
    return (itv_int_to_int(min_str), itv_int_to_int(max_str), modulo, warnings)
  if (itv_str[0] == '{'):
    # K-set: if it contains all values in the interval, we do not
    # generate a warning about precision
    # First we check if there is a "modulo" information
    warnings = []
    if "%" in itv_str:
      m = modulo_re.search(itv_str)
      mod_remainder = m.group(2)
      modulo = int(m.group(3))
      # We erase the modulo information from the string
      itv_str = m.group(1)
    itv_str = itv_str.strip("{}")
    separator = ";"
    elements = itv_str.split(separator)
    [min_str, max_str] = (elements[0], elements[-1])
    min_i = itv_int_to_int(min_str)
    max_i = itv_int_to_int(max_str)
    if max_i - min_i + 1 != len(elements):
      # Not all elements were inside
      warnings.append("interval [%s,%s] less precise than k-set %s" % (min_i, max_i, itv_str))
    return (min_i, max_i, modulo, warnings)
  if itv_str == "⊤" or itv_str == "T":
    # Top
    return (MIN_INT, MAX_INT, modulo, [])
  if itv_str == "Bot":
    return (MAX_INT, MIN_INT, modulo, [])
  raise Exception("Invalid interval: " + itv_str)

# Union between 2 intervals
def itv_union2(itv1, itv2):
  (m1, M1) = itv1
  (m2, M2) = itv2
  return (min(m1, m2),max(M1, M2))

# Union of a list of intervals
def itv_union(itvs):
  if len(itvs) == 0:
    return (MAX_INT, MIN_INT) # Bottom interval
  union = itvs[0]
  for i in range(1,len(itvs)):
    union = itv_union2(union, itvs[i])
  return union

# Warning: this function is not exact due to
# "modulo" boundary conditions.
# In practice we did not see any cases where
# it was relevant
def itv_len(i, modi):
  (a, b) = i
  # Special case: bottom
  if a > b:
    return 0
  else:
    return (b - a)/modi + 1

# The criterion for a bounded interval
def is_len_bounded(l):
  return l <= HALF_LENGTH

# Pretty-printinf of booleans
def yes_no(b):
  return "Y" if b else "N"

# Compute a union of intervals from a list of input lines
# Note: lines containing the same annotation line number must be
#       in consecutive order in the file, but they do not need to
#       be numerically sorted. Running the shell command 'sort' on 
#       the output is sufficient to ensure this condition.
def calculate_unions(lines):
  source_line_re = re.compile('^##(\d+): (\w+) = (.*)$')
  
  results = {}
  n = len(lines)
  [(line_number, var, itv_str)] = re.findall(source_line_re, lines[0])
  cur_line_number = line_number
  cur_var = var
  cur_union = [itv_str]
  i = 0
  while i < n:
    [(line_number, var, itv_str)] = re.findall(source_line_re, lines[i])
    i += 1
    if line_number == cur_line_number:
      # Same line: we merge the results
      #Sanity check
      if cur_var != var:
        raise Exception("Assertion error: same line_number %s, but different vars: %s != %s" % (line_number, cur_var, var))
      if itv_str in cur_union:
        # Interval already in union, nothing to do
        continue
      # Interval not yet in union
      cur_union.append(itv_str)
      continue
    else:
      # Not the same line: we store the previous union and start a new one
      results[(cur_line_number, cur_var)] = cur_union
      cur_line_number = line_number
      cur_var = var
      cur_union = [itv_str]
  # Then we save the last results
  results[(cur_line_number, cur_var)] = cur_union
  cur_line_number = line_number
  cur_var = var
  cur_union = [itv_str]
  return results

# Parses a string corresponding to a union of intervals,
# converting it into a list of intervals
def parse_union(union_str):
  union = []
  all_warnings = []
  lowest_modulo = 2**31
  for itv in union_str:
    (m, M, modulo, warnings) = itv_parse(itv)
    lowest_modulo = min(modulo, lowest_modulo)
    union.append((m, M))
    all_warnings.extend(warnings)
  return (union, lowest_modulo, all_warnings)

# Converts a list of intervals into a string
def union_to_str(union_str):
  res = ""
  if len(union_str) > 0:
    res += union_str[0]
  for i in union_str[1:]:
    res += " U " + i
  return res

# Returns whether a given variable name corresponds to a signed or unsigned 
# declared variable
def is_var_signed(var):
  # For now, we use the prefix "U_" to identify unsigned variables
  return var.strip()[0:2] != "U_"
