#!/bin/bash

usage() {
cat << EOF
usage: $0 [ -i <file.c> | -l <source_list.txt> ] [ -f | -c | -w ]+

Runs each of the specified analyzers on <file.c> (if option -i is used)
or on all files listed in <source_list.txt> (if option -l is used)

One result file is produced per analyzer.
Results are written to results.<analyzer>.txt for a file list,
or results.<file>.<analyzer>.txt for single files.

OPTIONS:
   -h                Display this message
   -i <file.c>       Define input C file
   -l <input.txt>    Define input file containing a list of C files
   -f                Run with Frama-C
   -c                Run with our value analysis
   -w                Run with Wrapped

EOF
}

run() {
  if [ "$FILES" == "" ]
  then
    echo "Error: input file(s) must be set before any analyzer"
    exit 1
  fi
  
  # The results file is name simply "results.<analyzer>.txt" if
  # a list has been used as input, otherwise it is
  # "results.<prog>.<analyzer>.txt".
  if [ "$FILE" == "" ]
  then
    # List as input
    resfile=results.${ANALYZER}.txt
  else
    BPROG=$(basename $FILE .c)
    resfile=results.${BPROG}.${ANALYZER}.txt
  fi
  
  echo -n "" > $resfile
  for f in $FILES
  do
    PROG=${f%.c}
    BPROG=$(basename $PROG)
    RES=${PROG}.${ANALYZER}.res
    echo "Running $f..."
    case $ANALYZER in
    fc)
      if [ ! -r "$SCR_DIR/framac_parse.sh" ]
      then
        echo "Warning: could not find script for Frama-C: $SCR_DIR/framac_parse.sh"
        echo "Ignoring flag -f, no output for Frama-C will be produced."
      else
        FRAMAC_FULL=${PROG}_framac_full.tmp
        $SCR_DIR/framac_parse.sh $f > $FRAMAC_FULL && grep "##" $FRAMAC_FULL | $SCR_DIR/pretty_limits.sh | sort > $RES
      fi
    ;;
    cc)
      ./cfgcomp.sh $f > $RES
    ;;
    wr)
      if [ ! -r "$SCR_DIR/wrapped_parse.sh" ]
      then
        echo "Warning: could not find script for Wrapped: $SCR_DIR/wrapped_parse.sh"
        echo "Ignoring flag -w, no output for Wrapped will be produced."
      else
        $SCR_DIR/wrapped_parse.sh $f > $RES
      fi
    ;;
    esac
    cat $RES # To provide a visual feedback to the user
    $SCR_DIR/count_bounded_dumps.sh $f $RES >> $resfile
  done

  # Pretty-print results
  ./print_results.py $resfile
}

# "main" starts here
SCR_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
did_something=0 # indicates if something has been done

while getopts "i:l:hfcw" OPTION
do
  case $OPTION in
  h)
    usage
    exit 1
  ;;
  i)
    FILE="$OPTARG"
    if [ ! -r $FILE ]
    then
      echo "Error: could not read input file: $FILE"
      exit 1
    fi
    FILES=$FILE
  ;;
  l)
    INPUT="$OPTARG"
    if [ ! -r $INPUT ]
    then
      echo "Error: could not read input list from file: $INPUT"
      exit 1
    fi
    FILES=$(cat $INPUT)
  ;;
  c)
    ANALYZER=cc
    run
    did_something=1
  ;;
  f)
    ANALYZER=fc
    run
    did_something=1
  ;;
  w)
    ANALYZER=wr
    run
    did_something=1
  ;;
  *)
    usage
    exit
  ;;
  esac
done

if [ "$did_something" = "0" ]
then
  usage
  exit
fi
