(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

open Printf
open Clflags
open Errors
open DriverUtil
open CFGAnalysis
open CFGDriverFlags
open IntervalDomain2aux (* contains command-line options *)
open BourdoncleIterator (* contains command-line options *)
open Benchs             (* contains command-line options *)

(* Location of the compatibility library *)

let stdlib_path = ref(
  try
    Sys.getenv "COMPCERT_LIBRARY"
  with Not_found ->
    Configuration.stdlib_path)

let command cmd =
  if !option_v then begin
    prerr_string "+ "; prerr_string cmd; prerr_endline ""
  end;
  Sys.command cmd

let quote_options opts =
  String.concat " " (List.rev_map Filename.quote opts)

let safe_remove file =
  try Sys.remove file with Sys_error _ -> ()

(* From C to preprocessed C *)

let preprocess ifile ofile =
  let output =
    if ofile = "-" then "" else sprintf "> %s" ofile in
  let cmd =
    sprintf "%s -D__COMPCERT__ %s %s %s %s"
      Configuration.prepro
      (if Configuration.need_stdlib_wrapper
       then sprintf "-I%s" !stdlib_path
       else "")
      (quote_options !prepro_options)
      ifile output in
  if command cmd <> 0 then begin
    if ofile <> "-" then safe_remove ofile;
    eprintf "Error during preprocessing.\n";
    exit 2
  end

(* From preprocessed C to Csyntax *)

let parse_c_file sourcename ifile =
  Sections.initialize();
  (* Simplification options *)
  let simplifs =
    "b" (* blocks: mandatory *)
  ^ (if !option_fstruct_return then "s" else "")
  ^ (if !option_fbitfields then "f" else "")
  ^ (if !option_fpacked_structs then "p" else "")
  in
  (* Parsing and production of a simplified C AST *)
  let ast =
    match Parse.preprocessed_file simplifs sourcename ifile with
    | None -> exit 2
    | Some p -> p in
  (* Remove preprocessed file (always a temp file) *)
  safe_remove ifile;
  (* Save C AST if requested *)
  if !option_dparse then begin
    let ofile = Filename.chop_suffix sourcename ".c" ^ ".parsed.c" in
    let oc = open_out ofile in
    Cprint.program (Format.formatter_of_out_channel oc) ast;
    close_out oc
  end;
  (* Conversion to Csyntax *)
  let csyntax =
    match C2C.convertProgram ast with
    | None -> exit 2
    | Some p -> p in
  flush stderr;
  (* Save CompCert C AST if requested *)
  if !option_dcmedium then begin
    let ofile = Filename.chop_suffix sourcename ".c" ^ ".compcert.c" in
    let oc = open_out ofile in
    PrintCsyntax.print_program (Format.formatter_of_out_channel oc) csyntax;
    close_out oc
  end;
  csyntax

(* Compiles the program with value analysis information *)
let compile_c_file_with_va sourcename ifile =
  let csyntax = parse_c_file sourcename ifile in
  (* Convert to CFG *)
  let cfg_prog = 
    if !option_no_inline then
      (* Do not inline the program *)
      do_res (DriverUtil.transf_c_program_to_cfg csyntax)
    else
      do_res (DriverUtil.transf_c_program_to_renumbered_cfg csyntax)
  in
  (* We print the value analysis results for the main function *)
  try
    let (main_id, main_f) = cfg_get_main cfg_prog in
    let _ = IntervalDomain2aux.function_int_constants main_f in
    if !option_dcfg then
      begin
      (* Print 'main' function to file *) 
        let filename = Filename.chop_suffix sourcename ".c" ^ ".main.cfg" in
        let oc = open_out filename in
        let pp = Format.formatter_of_out_channel oc in
        PrintCFG.print_function pp main_id main_f;
        close_out oc
      end;
    if !option_no_inline then
      begin
      (* Without inlining, we perform the value analysis on each function *)
        if !option_program_metrics then
          begin
            Benchs.print_program_metrics sourcename cfg_prog main_f None
          end;
        if !option_no_value then ()
        else
          PrintCFGValueAnalysis.print_value_analysis sourcename cfg_prog
      end
    else
      begin
        if !option_widening_hints > 0 then
          IntervalDomain2aux.update_widening_hints main_f;
        let o_tva = 
          if !option_no_value then None
          else 
            let (va_main, time_va) = Prof.time "Value_analysis" (fun _ -> Prof.timer (fun _ -> CFGAnalysis.vanalysis !option_numeric_domain main_f)) in
            Some (va_main, time_va)
        in
        if !option_program_metrics then
          begin
            Benchs.print_program_metrics sourcename cfg_prog main_f o_tva
          end
        else
          begin
            match o_tva with
              | None -> ()
              | Some (va_main, _) ->
                PrintCFGValueAnalysis.print_inlined_value_analysis sourcename main_id main_f va_main
          end
      end
  with
    | Not_found ->
      Printf.eprintf "%s: error: CFG value analysis requires a 'main' function.\n" sourcename;
      exit 2
        

(* Processing of a .c file *)

let process_c_file sourcename =
  let prefixname = Filename.chop_suffix sourcename ".c" in
  begin
    let preproname = Filename.temp_file "compcert" ".i" in
    preprocess sourcename preproname;
    compile_c_file_with_va sourcename preproname
  end;
  prefixname ^ ".out.cfg" (* Unused *)
    
(* Command-line parsing *)

let explode_comma_option s =
  match Str.split (Str.regexp ",") s with
  | [] -> assert false
  | hd :: tl -> tl

type action =
  | Set of bool ref
  | Unset of bool ref
  | Self of (string -> unit)
  | String of (string -> unit)
  | Integer of (int -> unit)
  | Unit of (unit -> unit)

let rec find_action s = function
  | [] -> None
  | (re, act) :: rem ->
      if Str.string_match re s 0 then Some act else find_action s rem

let parse_cmdline spec usage =
  let acts = List.map (fun (pat, act) -> (Str.regexp pat, act)) spec in
  let error () =
    eprintf "%s" usage;
    exit 2 in
  let rec parse i =
    if i < Array.length Sys.argv then begin
      let s = Sys.argv.(i) in
      match find_action s acts with
      | None ->
          if s <> "-help" && s <> "--help" 
          then eprintf "Unknown argument `%s'\n" s;
          error ()
      | Some(Set r) ->
          r := true; parse (i+1)
      | Some(Unset r) ->
          r := false; parse (i+1)
      | Some(Self fn) ->
          fn s; parse (i+1)
      | Some(String fn) ->
          if i + 1 < Array.length Sys.argv then begin
            fn Sys.argv.(i+1); parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; error()
          end
      | Some(Integer fn) ->
          if i + 1 < Array.length Sys.argv then begin
            let n =
              try
                int_of_string Sys.argv.(i+1)
              with Failure _ ->
                eprintf "Argument to option `%s' must be an integer\n" s;
                error()
            in
            fn n; parse (i+2)
          end else begin
            eprintf "Option `%s' expects an argument\n" s; error()
          end
      | Some (Unit f) -> f (); parse (i+1)
    end
  in parse 1

let usage_string =
"The Verified Value Analysis for CompCert version " ^ Configuration.version ^ "
Based on the CompCert C compiler.
Usage: cfgcomp [options] <source files>
Recognized source files:
  .c             C source file
Preprocessing options:
  -I<dir>        Add <dir> to search path for #include files
  -D<symb>=<val> Define preprocessor symbol
  -U<symb>       Undefine preprocessor symbol
  -Wp,<opt>      Pass option <opt> to the preprocessor
Language support options (use -fno-<opt> to turn off -f<opt>) :
  -fbitfields    Emulate bit fields in structs [off]
  -flonglong     Partial emulation of 'long long' types [on]
  -flongdouble   Treat 'long double' as 'double' [off]
  -fstruct-return  Emulate returning structs and unions by value [off]
  -fvararg-calls Emulate calls to variable-argument functions [on]
  -fpacked-structs  Emulate packed structs [off]
  -fall          Activate all language support options above
  -fnone         Turn off all language support options above
Tracing options:
  -dcfg          Save CFG of the main function (after inlining and renumbering)
                 in <file>.main.cfg
General options:
  -stdlib <dir>  Set the path of the Compcert run-time library [MacOS X only]
  -v             Print external commands before invoking them
  -metrics       Prints several metrics about the program
Value analysis options:
  -no-va         Disables the value analysis
  -no-inline     Do not inline functions; perform intraprocedural analysis
                 Note: this disables some options, such as -Widen/-WIDEN
  -dom-s         Numeric domain: signed intervals only [default if no option]
  -dom-u         Numeric domain: unsigned intervals only
  -dom-su        Numeric domain: reduced signed and unsigned
  -Widen         Use extra widening hints (slower analysis)
  -WIDEN         Use even more widening hints (even slower)
  -down          Number of down iterations (narrowing). Default: 1
"

let language_support_options = [
  option_fbitfields; option_flonglong; option_flongdouble;
  option_fstruct_return; option_fvararg_calls; option_fpacked_structs
]

let cmdline_actions =
  let f_opt name ref =
    ["-f" ^ name ^ "$", Set ref; "-fno-" ^ name ^ "$", Unset ref] in
  [
  "-I$", String(fun s -> prepro_options := s :: "-I" :: !prepro_options);
  "-D$", String(fun s -> prepro_options := s :: "-D" :: !prepro_options);
  "-U$", String(fun s -> prepro_options := s :: "-U" :: !prepro_options);
  "-[IDU].", Self(fun s -> prepro_options := s :: !prepro_options);
  "-stdlib$", String(fun s -> stdlib_path := s);
  "-dcfg$", Set option_dcfg;
  "-no-va$", Set option_no_value;
  "-no-inline$", Set option_no_inline;
  "-dom-s$", Unit(fun _ -> option_numeric_domain := SignedOnly);
  "-dom-u$", Unit(fun _ -> option_numeric_domain := UnsignedOnly);
  "-dom-su$", Unit(fun _ -> option_numeric_domain := SignedAndUnsignedReduced);
  "-Widen$", Unit(fun _ -> option_widening_hints := 1);
  "-WIDEN$", Unit(fun _ -> option_widening_hints := 2);
  "-down$", Integer(fun n -> option_down_iterations := n);
  "-metrics$", Set option_program_metrics;
  "-v$", Set option_v;
  ".*\\.c$", Self (fun s ->
      let objfile = process_c_file s in
      linker_options := objfile :: !linker_options);
  "-Wp,", Self (fun s ->
      prepro_options := List.rev_append (explode_comma_option s) !prepro_options);
  "-fsmall-data$", Integer(fun n -> option_small_data := n);
  "-fsmall-const$", Integer(fun n -> option_small_const := n);
  "-ffloat-const-prop$", Integer(fun n -> option_ffloatconstprop := n);
  "-falign-functions$", Integer(fun n -> option_falignfunctions := Some n);
  "-fall$", Self (fun _ ->
              List.iter (fun r -> r := true) language_support_options);
  "-fnone$", Self (fun _ ->
              List.iter (fun r -> r := false) language_support_options);
  "-debug.*", Self(fun s -> option_debug_options := List.tl (Str.split (Str.regexp " ") s); print_string "debug_options:"; List.iter (fun s' -> print_string " "; print_string s') !option_debug_options; print_newline ());
  "-print-va$", Set option_print_va_transfer;
  ]
  @ f_opt "longlong" option_flonglong
  @ f_opt "longdouble" option_flongdouble
  @ f_opt "struct-return" option_fstruct_return
  @ f_opt "bitfields" option_fbitfields
  @ f_opt "vararg-calls" option_fvararg_calls
  @ f_opt "packed-structs" option_fpacked_structs
  @ f_opt "sse" option_fsse

let _ =
  Gc.set { (Gc.get()) with Gc.minor_heap_size = 524288 };
  Machine.config :=
    begin match Configuration.arch with
    | "powerpc" -> Machine.ppc_32_bigendian
    | "arm"     -> Machine.arm_littleendian
    | "ia32"    -> Machine.x86_32
    | _         -> assert false
    end;
  Builtins.set C2C.builtins;
  CPragmas.initialize();
  parse_cmdline cmdline_actions usage_string
