(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** Pretty-printers for CFG code *)

open Format
open Camlcoq
open Datatypes
open Maps
open BinPos
open AST
open Integers
open Cminor
open CFG
open PrintAST
open PrintCminor
open Sort
open ToString

(* Printing of CFG code *)

let print_succ pp s dfl =
  let s = camlint_of_positive s in
  if s <> dfl then fprintf pp "       goto %ld@ " s

(* TODO: use a more readable version if the idents can be correctly associated
   to strings *)
let ident_name id =
  Printf.sprintf "<%ld>" (camlint_of_positive id)
  (*Printf.sprintf "%s<_%ld>" (PrintCminor.ident_name id) (camlint_of_positive id)*)

let print_optid pp = function
  | None -> ()
  | Some id -> fprintf pp "%s = " (ident_name id)

let rec ident_names_aux ids =
  match ids with
  | [] -> ""
  | e :: r -> ", " ^ ident_name e ^ ident_names_aux r

let ident_names ids =
  match ids with
  | [] -> ""
  | e :: r -> ident_name e ^ ident_names_aux r

let find_string_for_string_literal_id id =
  match Hashtbl.fold (fun s id' found0 -> if id = id' then s :: found0 else found0) C2C.stringTable []
  with
    | [] -> failwith (Printf.sprintf "Assertion error: no %s in stringTable" ~$id)
    | [s] -> s
    | _ -> failwith (Printf.sprintf "Assertion error: several matches for id %s in stringTable" ~$id)

let re_dump = Str.regexp "##.*"

(* Copied from PrintCminor, with less line breaks *)
let rec expr p (prec, e) =
  let (prec', assoc) = precedence e in
  let (prec1, prec2) =
    if assoc = LtoR
    then (prec', prec' + 1)
    else (prec' + 1, prec') in
  if prec' < prec 
  then fprintf p "("
  else fprintf p "";
  begin match e with
  | Evar id ->
      fprintf p "%s" (ident_name id)
  | Econst(Ointconst n) ->
      fprintf p "%ld" (camlint_of_coqint n)
  | Econst(Ofloatconst f) ->
      fprintf p "%F" (camlfloat_of_coqfloat f)
  | Econst(Oaddrsymbol(id, ofs)) ->
      let ofs = camlint_of_coqint ofs in
      if ofs = 0l
      then
        begin
          if Str.string_match PrintCsyntax.re_string_literal (extern_atom id) 0
          then
            (* String literal - we want to retrieve its "proper" value *)
            let s = find_string_for_string_literal_id id in
            if Str.string_match re_dump s 0
            then
              (* if it represents a DUMP, we print its value *)
              fprintf p "\"%s(\"%s\")\"" (extern_atom id) s
            else
              fprintf p "\"%s\"" (extern_atom id)
          else
            fprintf p "\"%s\"" (extern_atom id)
        end
      else fprintf p "(\"%s\" + %ld)" (extern_atom id) ofs
  | Econst(Oaddrstack n) ->
      fprintf p "&%ld" (camlint_of_coqint n)
  | Eunop(op, a1) ->
      fprintf p "%s %a" (name_of_unop op) expr (prec', a1)
  | Ebinop(op, a1, a2) ->
      fprintf p "%a %s %a"
                 expr (prec1, a1) (name_of_binop op) expr (prec2, a2)
  | Eload(chunk, a1) ->
      fprintf p "%s[%a]" (name_of_chunk chunk) expr (0, a1)
  | Econdition(a1, a2, a3) ->
      fprintf p "%a ? %a : %a" expr (4, a1) expr (4, a2) expr (4, a3)
  end;
  if prec' < prec then fprintf p ")" else fprintf p ""

(* Copied from PrintCminor, with less line breaks *)
let print_expr p e = expr p (0, e)

(* Copied from PrintCminor, with less line breaks *)
let rec print_expr_list p (first, rl) =
  match rl with
  | [] -> ()
  | r :: rl ->
      if not first then fprintf p ", ";
      expr p (2, r);
      print_expr_list p (false, rl)

let print_instruction pp (pc, i) =
  fprintf pp "%5ld: " pc;
  match i with
  | Iskip s ->
      let s = camlint_of_positive s in
      if s = Int32.pred pc
      then fprintf pp "skip@ "
      else fprintf pp "goto %ld@ " s
  | Iassign(id, e, s) ->
      fprintf pp "%s = %a@ " (ident_name id) print_expr e;
      print_succ pp s (Int32.pred pc)
  | Istore(chunk, addr, src, s) ->
      fprintf pp "%s[%a] = %a@ "
         (name_of_chunk chunk)
         print_expr addr
         print_expr src;
      print_succ pp s (Int32.pred pc)
  | Icall(optres, sg, fn, args, s) ->
      fprintf pp "%a%a(%a)@ "
        print_optid optres print_expr fn print_expr_list (true, args);
      print_succ pp s (Int32.pred pc)
  | Itailcall(sg, fn, args) ->
      fprintf pp "tailcall %a(%a)@ "
        print_expr fn print_expr_list (true, args)
  | Ibuiltin(optres, ef, args, s) ->
      fprintf pp "%a = builtin %s(%a)@ "
        print_optid optres (name_of_external ef) print_expr_list (true, args);
      print_succ pp s (Int32.pred pc)
  | Iifthenelse(cond, s1, s2) ->
      fprintf pp "if (%a) goto %ld else goto %ld@ "
        print_expr cond
        (camlint_of_positive s1) (camlint_of_positive s2)
  | Iswitch(a, cases, default) ->
      let tbl = Array.of_list cases in
      fprintf pp "@[<v 2>switch (%a)" print_expr a;
      for i = 0 to Array.length tbl - 1 do
        let (v, pc) = tbl.(i) in
        fprintf pp "@ case %ld: goto %ld" (camlint_of_z v) (camlint_of_positive pc)
      done;
      fprintf pp "@ default: goto %ld" (camlint_of_positive default);
      fprintf pp "@]@ "
  | Ireturn None ->
      fprintf pp "return@ "
  | Ireturn (Some arg) ->
      fprintf pp "return %a@ " print_expr arg

let print_function_header pp id f =
  fprintf pp "@[<v 2>%s(%s) {@ " (extern_atom id) (ident_names f.fn_params);
  fprintf pp "local vars: [%s]@ " (ident_names (List.sort cmp_pos_asc f.fn_vars))

(* Dumps the CFG version without value analysis *)
let print_function pp id f =
  print_function_header pp id f;
  let instrs =
    List.sort
      (fun (pc1, _) (pc2, _) -> Pervasives.compare pc2 pc1)
      (List.map
        (fun (pc, i) -> (camlint_of_positive pc, i))
        (PTree.elements f.fn_code)) in
  print_succ pp f.fn_entrypoint 
    (match instrs with (pc1, _) :: _ -> pc1 | [] -> -1l);
  List.iter (print_instruction pp) instrs;
  fprintf pp "@;<0 -2>}@]@."

let print_fundef pp (id, fd) =
  match fd with
  | Internal f -> print_function pp id f
  | External _ -> ()

let print_program pp (prog: CFG.program) =
  List.iter (print_fundef pp) prog.prog_funct

let print_if optdest prog =
  match !optdest with
  | None -> ()
  | Some f ->
      let oc = open_out f in
      let pp = formatter_of_out_channel oc in
      print_program pp prog;
      close_out oc

let destination_cfg : string option ref = ref None
let print_cfg = print_if destination_cfg

let destination_inlining_cfg : string option ref = ref None
let print_inlining_cfg = print_if destination_inlining_cfg
