(** Pretty-printers for the loop bound analysis, adding lots of debugging information. **)

open Printf
open Format
open Camlcoq
open Datatypes
open Errors
open Maps
open AST
open BinPos
open BinInt
open Integers
open Registers
open CfgBourdoncleAux
open Cminor
open LatticeSignatures
open CFG
open PrintAST
open PrintCminor
open CFGAnalysis
open DriverUtil
open ToString
open Sort
open IntervalDomain
open AbEnvironment
open Interval2ToString
open PrintCFG

(* TODO: should it be a command-line option? *)
let option_short_print = ref true

module PTree_Props = Tree_Properties(PTree)
let f_size f = camlint_of_nat (PTree_Props.cardinal f.fn_code)

type var = ident
type va_result = node -> (var -> sign_flag -> Interval.itv botlift) botlift
let dummy_var = Coq_xH

let rec calc_heights' (heights : (positive, int) Hashtbl.t) (b : B.bourdoncle) (cur_h : int) : unit =
  try
    match b with
      | B.I n -> Hashtbl.add heights n cur_h
      | B.L (h, l) -> 
        begin
          Hashtbl.add heights h (cur_h+1);
          List.iter (fun b' -> calc_heights' heights b' (cur_h+1)) l
        end
  with
    | Not_found -> failwith "Assertion error: height' not found!"

let calc_heights (f : coq_function) : (positive, int) Hashtbl.t =
  try
    let heights = Hashtbl.create (f_size f) in
    let (b, _) = build_bourdoncle f in
    calc_heights' heights b (-1);
    heights
  with
    | Not_found -> failwith "Assertion error: height not found!"

(* Iterval from its bounds
let itv_of (l, r) =
  { Interval.min = l; Interval.max = r }
*)

(* Returns true if p is a pair of two tops (gives no useful information) *)
let is_top_top p =
  let top_signed = 
    match p Signed with
      | Bot -> false
      | NotBot i -> i = IntervalDomain.Interval.top
  in
  let top_unsigned = 
    match p Unsigned with
      | Bot -> false
      | NotBot i -> i = IntervalDomain.Interval.utop
  in
  top_signed && top_unsigned

(* Printing of CFG code *)
let cat_itv s0 x = function p ->
  if is_top_top p then
    (* Print nothing *)
    s0
  else
    (* Print both intervals only if they differ *)
    let ss = itvbot_to_string (p Signed) in
    let su = itvbot_to_string (p Unsigned) in
    if ss = su then
      sprintf "%s %s:%s" s0 (ident_name x) ss        
    else
      sprintf "%s %s:<s%s|u%s>" s0 (ident_name x) ss su

let print_va pp (value : va_result) (fun_vars : node -> var list) (pc : positive) (i : instruction) =
  let va_str =
    begin
      match value pc with
      | Bot -> "Bot"
      | NotBot value_pc ->
            (* If not bottom, we print each variable *)
            List.fold_left (fun s0 x -> cat_itv s0 x (value_pc x))
            ""
            (fun_vars pc)
    end
  in
  fprintf pp "    %s\n" va_str

let print_instruction pp (value : va_result) (fun_vars : node -> var list) heights get_prefix (pc, i) =
  let ppc = Camlcoq.positive_of_camlint pc in
  let height_pc = 
    try 
      Hashtbl.find heights ppc
    with
      | Not_found -> 
        begin
          if ppc != Coq_xH then Printf.eprintf "Warning: could not find height for %ld (unreachable?)\n" pc;
          0
        end
  in
  let spaces = String.make (4 * height_pc) ' ' in
  fprintf pp "%s" spaces;
  print_va pp value fun_vars ppc i;
  fprintf pp "%c%s" (get_prefix pc) spaces;
  fprintf pp "%4ld: " pc;
  match i with
  | Iskip s ->
      let s = camlint_of_positive s in
      if s = Int32.pred pc
      then fprintf pp "skip@ "
      else fprintf pp "goto %ld@ " s
  | Iassign(id, e, s) ->
      fprintf pp "%s = %a@ " (ident_name id) print_expr e;
      print_succ pp s (Int32.pred pc)
  | Istore(chunk, addr, src, s) ->
      fprintf pp "%s[%a] = %a@ "
         (name_of_chunk chunk)
         print_expr addr
         print_expr src;
      print_succ pp s (Int32.pred pc)
  | Icall(optres, sg, fn, args, s) ->
      fprintf pp "%a%a(%a)@ "
        print_optid optres print_expr fn print_expr_list (true, args);
      print_succ pp s (Int32.pred pc)
  | Itailcall(sg, fn, args) ->
      fprintf pp "tailcall %a(%a)@ "
        print_expr fn print_expr_list (true, args)
  | Ibuiltin(optres, ef, args, s) ->
      fprintf pp "%a = builtin %s(%a)@ "
        print_optid optres (name_of_external ef) print_expr_list (true, args);
      print_succ pp s (Int32.pred pc)
  | Iifthenelse(cond, s1, s2) ->
      fprintf pp "if (%a) goto %ld else goto %ld@ "
        print_expr cond
        (camlint_of_positive s1) (camlint_of_positive s2)
  | Iswitch(a, cases, default) ->
      let tbl = Array.of_list cases in
      fprintf pp "@[<v 2>switch (%a)" print_expr a;
      for i = 0 to Array.length tbl - 1 do
        let (v, pc) = tbl.(i) in
        fprintf pp "@ case %ld: goto %ld" (camlint_of_z v) (camlint_of_positive pc)
      done;
      fprintf pp "@ default: goto %ld" (camlint_of_positive default);
      fprintf pp "@]@ "
  | Ireturn None ->
      fprintf pp "return@ "
  | Ireturn (Some arg) ->
      fprintf pp "return %a@ " print_expr arg

let print_function pp id f heights (get_prefix : int32 -> char) (value : va_result) =
  print_function_header pp id f;
  let vars = List.sort cmp_pos_asc (f.fn_params @ f.fn_vars) in
  let fun_vars = fun _node -> vars in
  let instrs =
    List.sort
      (fun (pc1, _) (pc2, _) -> Pervasives.compare pc2 pc1)
      (List.map
        (fun (pc, i) -> (camlint_of_positive pc, i))
        (PTree.elements f.fn_code)) in
  print_succ pp f.fn_entrypoint 
    (match instrs with (pc1, _) :: _ -> pc1 | [] -> -1l);
  List.iter (print_instruction pp value fun_vars heights get_prefix) instrs;
  fprintf pp "@;<0 -2>}@]@."

(* 'value' already contains the result of the analysis *)
let print_inlined_value_analysis sourcename id f value =
  let filename = Filename.chop_suffix sourcename ".c" ^ ".value.cfg" in
  let oc = open_out filename in
  let pp = formatter_of_out_channel oc in
  let heights = calc_heights f in
  print_function pp id f heights (fun _ -> ' ') (* no slice: prefix = ' ' *) value;
  close_out oc

(* This version computes the value analysis for each function *)
let print_value_analysis sourcename prog =
  let filename = Filename.chop_suffix sourcename ".c" ^ ".no_inline.value.cfg" in
  let oc = open_out filename in
  let pp = formatter_of_out_channel oc in
  List.iter (fun (id, fd) ->
    match fd with
      | Internal f -> 
        begin
          let (value, time) = Prof.timer (fun _ -> CFGAnalysis.vanalysis !CFGDriverFlags.option_numeric_domain f) in
          let heights = calc_heights f in
          print_function pp id f heights (fun _ -> ' ') (* no slice: prefix = ' ' *) value;
          if !CFGDriverFlags.option_program_metrics then
            Printf.printf ">>>%s: Value analysis time for function %s: %f\n" sourcename (extern_atom id) time;
        end
      | External _ -> ()
  ) prog.prog_funct;
  close_out oc
