open Camlcoq
open Datatypes
open Maps
open Integers
open AST
open Cminor
open CFG
open Printf
open ToString

let cfg_get_function (prog : program) (id : ident) : coq_function option =
  try 
    match List.assoc id prog.prog_funct with
      | External _ -> None
      | Internal f -> Some f
  with
    | Not_found -> failwith (sprintf "could not find function: %s" (extern_atom id))

module PTree_Props = Tree_Properties(PTree)
let f_size f = camlint_of_nat (PTree_Props.cardinal f.fn_code)

let get_internal_function_ids (prog : program) : ident list =
  List.map fst (List.filter (fun (id, fd) -> match fd with Internal _ -> true | _ -> false) prog.prog_funct)

let print_cfg_function_size p (f : coq_function) : unit =
  p (sprintf "CFG instructions: %d" (f_size f))

(* NOTE: this must be udpated whenever debugging function names change *)
let debug_function_names = ["malloc"; "free"; "realloc"; "calloc"; "__dump"; "nondet_int"; "__builtin_nondet_int"]

let remove_debugging_functions (ids : ident list) : ident list =
  List.filter (fun id -> not (List.mem (extern_atom id) debug_function_names)) ids

let print_count_program_functions p (prog : program) : unit =
  p (sprintf "Number of functions: %d" (List.length (prog.prog_funct)));
  let count_int_fs = List.length (List.filter (fun (id, fd) -> match fd with Internal _ -> true | _ -> false) prog.prog_funct) in
  p (sprintf "Number of internal functions: %d" count_int_fs);
  let internal_ids = get_internal_function_ids prog in
  p (sprintf "Names of internal functions: %s" (listToString extern_atom internal_ids));
  let good_ids =  remove_debugging_functions internal_ids in
  p (sprintf "Number of internal functions (without dumps and debug): %d" (List.length good_ids))

let cfg_memoized_functions : (ident, int) Hashtbl.t = Hashtbl.create 16

(* Note: This computes an under-approximation: calls using function
         pointers are not taken into account.
   Returns a pair containing the size of the argument list (to estimate the number
   of additional nodes that will be created by the inlining) and a function
   identifier, to recursively estimate its size. *)
let cfg_get_called_function_ids (f : coq_function) : (int * ident) list =
  PTree.fold (fun cs0 _p i ->
    match i with
      | Icall (_optres, _sg, fn, args, _s) -> 
        begin
          match fn with
            | Econst c ->
              begin
                match c with
                  | Oaddrsymbol(id, offset) ->
                    if Int.eq offset Int.zero then
                      (List.length args, id) :: cs0
                    else
                      cs0
                  | _ -> cs0
              end
            | _ -> cs0
        end
      | Itailcall(_sg, fn, args) ->
        begin
          match fn with
            | Econst c ->
              begin
                match c with
                  | Oaddrsymbol(id, offset) ->
                    if Int.eq offset Int.zero then
                      (List.length args, id) :: cs0
                    else
                      cs0
                  | _ -> cs0
              end
            | _ -> cs0
        end
      | _ -> cs0
  ) f.fn_code []

(* The call stack prevents infinite recursion *)
let rec cfg_recursive_count_nodes (id_f : ident -> coq_function option) (call_stack : ident list) (cur_id : ident) (f : coq_function) : int =
  try 
    Hashtbl.find cfg_memoized_functions cur_id
  with
    | Not_found ->
      let cargs_ids = cfg_get_called_function_ids f in
      (* Finds and prints recursive calls *)
      let recursive_ids = List.filter (fun (cargs, i) -> List.mem i call_stack) cargs_ids in
      if recursive_ids <> [] then
        Printf.printf "recursive call detected: call stack: %s, ids: %s\n" (listToString extern_atom call_stack) (listToString extern_atom (cur_id :: List.map snd recursive_ids));
      let non_recursive_cargs_ids = List.filter (fun (cargs, i) -> not (List.mem i call_stack)) cargs_ids in
      let size_children = 
        List.fold_left (fun c0 (cargs, i) -> 
          match id_f i with
            | None -> c0
            | Some f -> cargs + c0 + (cfg_recursive_count_nodes id_f (cur_id :: call_stack) i f)
        ) 0 non_recursive_cargs_ids
      in
      let size = f_size f + size_children in
      Hashtbl.add cfg_memoized_functions cur_id size;
      size

(* Returns a lower estimate for the number of nodes in the inlined 
   program *)
let cfg_estimate_inlined_nodes (cfg_prog : program) : int =
  let (main_id, main_f) = DriverUtil.cfg_get_main cfg_prog in
  let id_f = (fun id -> cfg_get_function cfg_prog id) in
  let count = cfg_recursive_count_nodes id_f [] main_id main_f in
  printf "INLINED_NODES : %d\n" count;
  count

let print_program_metrics (name : string) (prog : program) (main_f : coq_function) o_tva =
  printf "*** Program metrics for file: %s ***\n" name;
  let printer = fun s -> printf ">>>%s: %s\n" name s in
  begin
    match o_tva with
      | None -> ()
      | Some (va, time) ->
        printer (sprintf "value analysis time: %f" time)
  end;
  print_cfg_function_size printer main_f;
  print_count_program_functions printer prog;
  () (* end print_program_metrics *)




(*************** RTL functions ***************)

open RTL

let get_function (prog : program) (id : ident) : coq_function option =
  try 
    match List.assoc id prog.prog_funct with
      | External _ -> None
      | Internal f -> Some f
  with
    | Not_found -> failwith (sprintf "could not find function: %s" (extern_atom id))

(* Note: This computes an under-approximation: calls using function
         pointers are not taken into account. *)
let get_called_function_ids (f : coq_function) : (int * ident) list =
  PTree.fold (fun cs0 _p i ->
    match i with
      | Icall(_sig0, ros, args, _res, _s) -> 
        begin
          match ros with
            | Coq_inl r -> cs0
            | Coq_inr id -> (List.length args, id) :: cs0
        end
      | Itailcall(_sig0, ros, args) ->
        begin
          match ros with
            | Coq_inl r -> cs0
            | Coq_inr id -> (List.length args, id) :: cs0
        end
      | _ -> cs0
  ) f.fn_code []

let memoized_functions : (ident, int) Hashtbl.t = Hashtbl.create 16

let f_size f = camlint_of_nat (PTree_Props.cardinal f.fn_code)

(* The call stack prevents infinite recursion *)
let rec recursive_count_nodes (id_f : ident -> coq_function option) (call_stack : ident list) (cur_id : ident) (f : coq_function) : int =
  try 
    Hashtbl.find memoized_functions cur_id
  with
    | Not_found ->
      let cargs_ids = get_called_function_ids f in
      let non_recursive_cargs_ids = List.filter (fun (cargs, i) -> not (List.mem i call_stack)) cargs_ids in
      let size_children = 
        List.fold_left (fun c0 (cargs, i) -> 
          match id_f i with
            | None -> c0
            | Some f -> cargs + c0 + (recursive_count_nodes id_f (cur_id :: call_stack) i f)
        ) 0 non_recursive_cargs_ids
      in
      let size = f_size f + size_children in
      Hashtbl.add memoized_functions cur_id size;
      size

(* Returns a lower estimate for the number of nodes in the inlined 
   program *)
let estimate_inlined_nodes (rtl_prog : program) : int =
  let (main_id, main_f) = DriverUtil.get_main rtl_prog in
  let id_f = (fun id -> get_function rtl_prog id) in
  let count = recursive_count_nodes id_f [main_id] main_id main_f in
  printf "INLINED_NODES : %d\n" count;
  count
