(* *********************************************************************)
(*                                                                     *)
(*              The Compcert verified compiler                         *)
(*                                                                     *)
(*          Xavier Leroy, INRIA Paris-Rocquencourt                     *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique.  All rights reserved.  This file is distributed       *)
(*  under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 2 of the License, or  *)
(*  (at your option) any later version.  This file is also distributed *)
(*  under the terms of the INRIA Non-Commercial License Agreement.     *)
(*                                                                     *)
(* *********************************************************************)

(** Pretty-printer for Cminor *)

open Format
open Camlcoq
open Datatypes
open BinPos
open Integers
open AST
open PrintAST
open Cminor
open PrintCminor
open AbMachineNumSignature

(* Expressions *)

let option_use_ident_name = true

let ident_name id =
  let s = 
    match id with
      | Coq_xO n -> extern_atom n
      | Coq_xI n -> Printf.sprintf "$%ld" (camlint_of_positive n)
      | Coq_xH -> "$0"
  in
  if option_use_ident_name then
    Printf.sprintf "<%s>{%s}" (ToString.(~$) id) s
  else
    Printf.sprintf "<%s>" (ToString.(~$) id)

let rec expr_of_nexpr : 'a nexpr -> expr = function
  | NEvar v -> Evar Coq_xH
  | NEconst (NOintconst i) -> Econst (Ointconst i)
  | NEunop (op, e) -> Eunop (op, expr_of_nexpr e)
  | NEbinop (op, e, f) -> Ebinop (op, expr_of_nexpr e, expr_of_nexpr f)
  | NEcondition (b, e, f) -> Econdition (expr_of_nexpr b, expr_of_nexpr e, expr_of_nexpr f)
  | NEconst NOintunknown -> Econst (Ointconst (coqint_of_camlint 42l))

let rec expr (prec, e) =
  let (prec', assoc) = precedence e in
  let (prec1, prec2) =
    if assoc = LtoR
    then (prec', prec' + 1)
    else (prec' + 1, prec') in
  let str_1 = 
    if prec' < prec 
    then sprintf "("
    else sprintf ""
  in
  let str_2 = 
    match e with
      | Evar id ->
        sprintf "Evar(%s)" (ident_name id)
      | Econst(Ointconst n) ->
        sprintf "Econst(%ld)" (camlint_of_coqint n)
      | Econst(Ofloatconst f) ->
        sprintf "Econst(%F)" (camlfloat_of_coqfloat f)
      | Econst(Oaddrsymbol(id, ofs)) ->
        let ofs = camlint_of_coqint ofs in
        if ofs = 0l
        then sprintf "Econst(\"%s\")" (extern_atom id)
        else sprintf "Econst(\"%s\" + %ld)" (extern_atom id) ofs
      | Econst(Oaddrstack n) ->
        sprintf "Econst(addrstack(&%ld))" (camlint_of_coqint n)
      | Eunop(op, a1) ->
        sprintf "%s %s" (name_of_unop op) (expr (prec', a1))
      | Ebinop(op, a1, a2) ->
        sprintf "%s %s %s"
          (expr (prec1, a1)) (name_of_binop op) (expr (prec2, a2))
      | Eload(chunk, a1) ->
        sprintf "%s[%s]" (name_of_chunk chunk) (expr (0, a1))
      | Econdition(a1, a2, a3) ->
        sprintf "%s ? %s : %s" (expr (4, a1)) (expr (4, a2)) (expr (4, a3))
  in
  let str_3 = if prec' < prec then sprintf ")" else sprintf ""
  in
  str_1 ^ str_2 ^ str_3

let dump_expr e = expr (0, e)

let dump_nexpr e = dump_expr (expr_of_nexpr e)
