(* Heuristics for interval widening *)

open BinPos
open BinInt
open Camlcoq
open Integers
open Maps
open Cminor
open CFG
open Sort
open ToString

let option_widening_hints = ref 0

(* mutable set *)
type mset = (Int.int, unit) Hashtbl.t

let add s i = Hashtbl.replace s i ()

let add_all (s : mset) (l : Int.int list) =
  List.iter (fun i -> Hashtbl.add s i ()) l

let elements (s : mset) : Int.int array =
  let t = Array.make (Hashtbl.length s) Int.zero in
  let _ = Hashtbl.fold (fun e _ i -> t.(i) <- e; i+1) s 0 in
  t

let rec add_expr_int_constants (m : mset) (e : expr) : unit =
  let aux = add_expr_int_constants m in
  match e with
    | Econst(Ointconst n) -> add m n
    | Eunop(op, a1) -> aux a1
    | Ebinop(op, a1, a2) -> aux a1; aux a2
    | Econdition(a1, a2, a3) -> aux a1; aux a2; aux a3
    | _ -> ()

let add_instr_int_constants (m : mset) (i : instruction) : unit =
  match i with
    | Iskip s -> ()
    | Iassign(id, e, s) -> add_expr_int_constants m e
    | Istore(chunk, addr, src, s) -> ()
    | Icall(optres, sg, fn, args, s) -> List.iter (add_expr_int_constants m) args
    | Itailcall(sg, fn, args) -> List.iter (add_expr_int_constants m) args
    | Ibuiltin(optres, ef, args, s) -> List.iter (add_expr_int_constants m) args
    | Iifthenelse(cond, s1, s2) -> add_expr_int_constants m cond
    | Iswitch(a, cases, default) -> add_expr_int_constants m a
    | _ -> ()

let function_int_constants (f : coq_function) : mset =
  let constants = Hashtbl.create 7 in
  PTree.fold (fun t0 p i -> add_instr_int_constants t0 i; t0) f.fn_code constants
  
(* Default constants always used as hints *)
let default_hints : Int.int list = List.map z_of_camlint [-65535l; -256l; -128l; -1l; 1l; 128l; 256l; 65536l]

(* Array of widening hints - must be built using function_int_constants *)
let hints : Int.int array ref = ref (Array.make 0 Int.zero)

(* Adds hints for "-1" and "+1" cases *)
let add_border_hints (hints : mset) : unit =
  Hashtbl.iter (fun h _ -> add hints (coq_Zpred h); add hints h; add hints (coq_Zsucc h)) hints

let update_widening_hints (main_f : coq_function) : unit =
  let constants = function_int_constants main_f in
  add_all constants default_hints;
  if !option_widening_hints >= 2 then
    add_border_hints constants;
  let t = elements constants in
  Array.sort cmp_z_asc t;
  Printf.printf "Using %d widening hints\n" (Array.length t);
  hints := t


let rec next_pow2_pos = function
  | Coq_xI p0 -> Coq_xI (next_pow2_pos p0)
  | Coq_xO p0 -> Coq_xI (next_pow2_pos p0)
  | Coq_xH -> Coq_xH
    
let previous_pow2_pos p =
  let p' = next_pow2_pos p in
  (match p' with
    | Coq_xI p0 -> p0
    | _ -> p')

let next_threshold (z : coq_Z) = 
  match z with
    | Z0 -> Z0
    | Zpos Coq_xH -> Zpos Coq_xH
    | Zneg Coq_xH -> Zneg Coq_xH
    | Zpos p -> Zpos (next_pow2_pos p)
    | Zneg p -> Zneg (previous_pow2_pos p)

let previous_threshold (z : coq_Z) = 
  match z with
    | Z0 -> Z0
    | Zpos Coq_xH -> Zpos Coq_xH
    | Zneg Coq_xH -> Zneg Coq_xH
    | Zpos p -> Zpos (previous_pow2_pos p)
    | Zneg p -> Zneg (next_pow2_pos p)

exception Found of coq_Z

let next_larger (z : coq_Z) : coq_Z =
  let next_t = next_threshold z in
  if !option_widening_hints <= 0 then
    next_t
  else
    begin
      let res = 
        let next_t = next_threshold z in
        let n = Array.length !hints - 1 in
        try 
          for i = 0 to n do
            if cmp_z_asc z !hints.(i) < 0 then
              begin
                (* found someone larger than z *)
                raise (Found !hints.(i));
              end
          done;
          (* default if none found *)
          next_t
        with
            Found f -> 
              (* we choose the best approximation *)
              Zminmax.coq_Zmin next_t f
      in
      res
    end

let next_smaller (z : coq_Z) : coq_Z =
  let previous_t = previous_threshold z in
  if !option_widening_hints <= 0 then
    previous_t
  else
    begin
      let res =
        let n = Array.length !hints - 1 in
        try 
          for i = n downto 0 do
            if cmp_z_asc z !hints.(i) > 0 then
              begin
                (* found someone smaller than z *)
                raise (Found !hints.(i));
              end
          done;
          (* default if none found *)
          previous_t
        with
            Found f -> 
              (* we choose the best approximation *)
              Zminmax.coq_Zmax previous_t f
      in
      res
    end
